//-----------------------------------------------------------------------------
// EXT_LCD_EADOG_ST7565
// Version 1.0
// 22.11.2012 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer ein externes LC-Display
// Typ          : EA-DOG-Display
// Display-Chip : ST7565R
// Modell       : z.B. EA-DOGM_132_5
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_EXT_MODULE/"
// liegen und mit #include <UB_EXT_MODULE/EXT_LCD_EADOG_ST7565.h>
// eingebunden werden
//          
//----------------------------------------------------------------------------- 
// Belegt : Port-B (Bit4,5)
// Benutzt : MK3_2560_SPI.h = SPI = Port-B (Bit1 bis Bit3)
//-----------------------------------------------------------------------------
//
// Patchplan : (fuer den Mode : 4Wire, 8bit, SPI)
//
// (in der Default-Einstellung liegt alles auf Port-B)
//
// Port-B1 (SPI-SCK)  muss zum Display Pin SCL
// Port-B2 (SPI-MOSI) muss zum Display Pin SI
// Port-B4 (A0) muss zum Display Pin A0
// Port-B5 (CS) muss zum Display Pin CS1
//
//-----------------------------------------------------------------------------



//----------------------------------------------------------------------------
// Includes
//---------------------------------------------------------------------------- 

#include <stdio.h>
#include "ub_mk3_2560/MK3_2560_SPI.h" //  bindet die SPI-Funktionen ein
#include <avr\pgmspace.h>



//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------

#define  ST7565_DDR    DDRB
#define  ST7565_PORT   PORTB
#define  ST7565_A0     0x10    // Bit4
#define  ST7565_CS     0x20    // Bit5


//-----------------------------------------------------------------------------
// Display Groesse
// (Bei EA-DOG 132-5 -> Pixel_X=132, Pixel_Y=32, Pages_Y=4)
//----------------------------------------------------------------------------
#define  ST7565_PIXEL_X  132   // Anzahl der Pixel in X-Richtung
#define  ST7565_PIXEL_Y  32    // Anzahl der Pixel in Y-Richtung
#define  ST7565_PAGES_Y  4     // Anzahl der Pages in Y-Richtung (1 Page = 8 Pixel)   

//-----------------------------------------------------------------------------
// Display Befehle
//-----------------------------------------------------------------------------
#define  ST7565_SET_COLUMN_ADR_L  0x00     // Zeilen-Adr Lo
#define  ST7565_SET_COLUMN_ADR_H  0x10     // Zeilen-Adr Hi
#define  ST7565_SET_PAGE_ADR      0xB0     // Page-Adr
#define  ST7565_DATA_TO_LCD       0xFF     // Daten zum Display



//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//----------------------------------------------------------------------------- 
void EXT_LCD_INIT_ST7565(void);                       // init vom LC-Display
void LCD_ST7565_CLEAR(void);                          // loescht das Display
void LCD_ST7565_SETCURSOR(char x, char y);            // setzt den Cursor
void LCD_ST7565_PRINT_BYTE(unsigned char wert);       // zeichnet ein Byte ab Cursor
void LCD_ST7565_PRINT_ASC(unsigned char wert);        // zeichnet ein Ascii ab Cursor
void LCD_ST7565_PRINT_UINT(unsigned int wert);        // schreibt eine UINT ab Cursor
void LCD_ST7565_STRING(const char *ptr);              // schreibt einen String ab Cursor
void LCD_ST7565_STRING_AT(char x,char y,const char *ptr);  // schreibt einen String



//-----------------------------------------------------------------------------
// interne Variabeln
//-----------------------------------------------------------------------------

typedef enum {
  ST7565_CMD1 =0, // Command-Befehl fuer das Display (1 Byte Befehl)
  ST7565_CMD2,    // Command-Befehl fuer das Display (Mehrbyte Befehl)
  ST7565_DATA     // Data-Befehl fuer das Display (1 Byte Befehl)
}ST7565_CMD_t;

typedef struct {
  unsigned char akt_page_adr; // aktuelle Page Adresse
  unsigned int akt_col_adr;   // aktuelle Zeilen Adresse
}ST7565_LCD_t;
ST7565_LCD_t ST7565_LCD;



//-----------------------------------------------------------------------------
// interne Funktionen
//-----------------------------------------------------------------------------
void ST7565_initseq(void);
void ST7565_send(ST7565_CMD_t mode, unsigned char wert);
void ST7565_sendCommand(char befehl, char wert);



//-----------------------------------------------------------------------------
// hier wird ein eigener FLASH-Bereich definiert,damit keine
// Compiler-Warning erzeugt wird, wie beim benutzen von "PROGMEM"
//-----------------------------------------------------------------------------
#define UB_PROGMEM __attribute((section(".progmem.ubflash")))


//-----------------------------------------------------------------------------
// Init vom LC-Display
//-----------------------------------------------------------------------------
void EXT_LCD_INIT_ST7565(void) { // init vom LC-Display
  // Init vom SPI-Bus
  MK3_INIT_SPI();

  // SPI-Einstellungen
  // CPOL=1  CPHA=1
  MK3_SPI_SET_MODE(SPI_MODE_3); 

  //  Teiler=128
  MK3_SPI_SET_FRQ(SPI_CLK_DIV_128);

  // Display-Pins auf Ausgang
  ST7565_DDR|=ST7565_A0;
  ST7565_DDR|=ST7565_CS;

  // Display-Pins auf Hi-Pegel
  ST7565_PORT|=ST7565_A0;
  ST7565_PORT|=ST7565_CS;

  // init der Variabeln
  ST7565_LCD.akt_page_adr=0;
  ST7565_LCD.akt_col_adr=0;

  // Init-Sequenz vom Display starten
  ST7565_initseq();
}

//-----------------------------------------------------------------------------
// loescht das Display komplett
//-----------------------------------------------------------------------------
void LCD_ST7565_CLEAR(void)  // loescht das Display
{   
  unsigned char x,y;  
  
  // Cursor auf 0 setzen 
  
  // alle Pages durchgehen  
  for(y=0;y<ST7565_PAGES_Y;y++) {
    ST7565_LCD.akt_col_adr=0;
    ST7565_sendCommand(ST7565_SET_COLUMN_ADR_L,0);
    ST7565_sendCommand(ST7565_SET_COLUMN_ADR_H,0);

    ST7565_sendCommand(ST7565_SET_PAGE_ADR,y);
        
    // alle Spalten durchgehen
    for(x=1;x<=ST7565_PIXEL_X;x++) {
      ST7565_sendCommand(ST7565_DATA_TO_LCD,0x00);  
    }  
 } 
}

//-----------------------------------------------------------------------------
// setzt den Cursor an x,y position
// x = X-Position in Pixel
// y = Y-Position als Page
//-----------------------------------------------------------------------------
void LCD_ST7565_SETCURSOR(char x, char y) { // setzt den Cursor
  unsigned char xl,xh;

  // test auf grenzen  
  if(x>=ST7565_PIXEL_X) x=(ST7565_PIXEL_X-1);
  if(y>=ST7565_PAGES_Y) y=(ST7565_PAGES_Y-1);
  
  // y-pos setzen
  ST7565_sendCommand(ST7565_SET_PAGE_ADR,y);   
  
  ST7565_LCD.akt_col_adr=x;  
  xl=(char)(x&0x0F);
  xh=(char)(x>>4);    
  ST7565_sendCommand(ST7565_SET_COLUMN_ADR_L,xl);
  ST7565_sendCommand(ST7565_SET_COLUMN_ADR_H,xh);  
}

//-----------------------------------------------------------------------------
// zeichnet ein Byte an Cursor-Pos
//-----------------------------------------------------------------------------
void LCD_ST7565_PRINT_BYTE(unsigned char wert) { // zeichnet ein Byte
  ST7565_sendCommand(ST7565_DATA_TO_LCD,wert);
}

//-----------------------------------------------------------------------------
// schreibt einen String an Cursor-Pos
//-----------------------------------------------------------------------------
void LCD_ST7565_STRING(const char *ptr) { // schreibt einen String ab cursor
  unsigned int n;
  unsigned char c;

  for(n=0;n<=255;n++) {
    c=ptr[n];	
    if(c==0) {
      break; 
    }
    else {
      LCD_ST7565_PRINT_ASC(c);
    }   
  } 
}

//-----------------------------------------------------------------------------
// schreibt einen String
// an eine bestimmte Position
//-----------------------------------------------------------------------------
void LCD_ST7565_STRING_AT(char x,char y,const char *ptr) { // schreibt einen String  

  // cursor setzen
  LCD_ST7565_SETCURSOR(x,y);

  // string schreiben
  LCD_ST7565_STRING(ptr);
}

//-----------------------------------------------------------------------------
// schreibt eine UINT an Cursor-Pos
//-----------------------------------------------------------------------------
void LCD_ST7565_PRINT_UINT(unsigned int wert) { // schreibt eine UINT
  char buf[7];

  // Zahl in String wandeln
  sprintf(buf,"%5u",wert); // 5 stellige unsigned_int (ohne fuehrenden Nullen)

  // String ausgeben
  LCD_ST7565_STRING(buf);
}

//-----------------------------------------------------------------------------
// Init-Sequenz (LOW-POWER : siehe Datenblatt)
//-----------------------------------------------------------------------------
void ST7565_initseq(void) { // interne Funktion
  unsigned int n,m;  
  
  // kleine Pause (WICHTIG !!)
  for(n=0;n<100;n++) {
    for(m=0;m<1000;m++) {
      // kleine Pause
    }
  }

  // setze Display Startlinie auf 0
  // Befehl Nr. 2 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0x40); 

  // setze ADC-Mode auf REVERSE
  // Befehl Nr. 8 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0xA1);  

  // lcd mapping auf NORMAL
  // Befehl Nr. 15 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0xC0);

  // setze Display auf NORMAL
  // Befehl Nr. 9 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0xA6);

  // setze BIAS auf 1/9
  // Befehl Nr. 11 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0xA2);

  // setze Power (Booster,Regulator,Follower)
  // Befehl Nr. 16 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0x2F);

  // setze Booster auf 3x/4x 
  // Befehl Nr. 20 (2Byte Befehl)
  ST7565_send(ST7565_CMD2,0xF8);
  ST7565_send(ST7565_CMD1,0x00);

  // setze V0 Voltage
  // Befehl Nr. 17 (1Byte Befehl)
  ST7565_send(ST7565_CMD1,0x23);

  // setze Contrast auf 1Fh = 31d 
  // Befehl Nr. 18 (2Byte Befehl)
  ST7565_send(ST7565_CMD2,0x81);
  ST7565_send(ST7565_CMD1,0x1F);

  // setze Indicator auf OFF
  // Befehl Nr. 19 (2Byte Befehl)
  ST7565_send(ST7565_CMD2,0xAC);
  ST7565_send(ST7565_CMD1,0x00);

  // schalte Display EIN
  // Befehl Nr. 1
  ST7565_send(ST7565_CMD1,0xAF);   
}



//-----------------------------------------------------------------------------
// sendet einen Befehl an das Display
//-----------------------------------------------------------------------------
void ST7565_send(ST7565_CMD_t mode, unsigned char wert) { // interne Funktion 
  if((mode==ST7565_CMD1) || (mode==ST7565_CMD2)) {
    // mode auf control
    ST7565_PORT&=~ST7565_A0;  
  }
  else {
    // mode auf data
    ST7565_PORT|=ST7565_A0;
    ST7565_LCD.akt_col_adr+=1;    
  }
   
  // Chip-Select auf Lo
  ST7565_PORT&=~ST7565_CS;
   
  //---------------------------------------------------
  //---- Byte per SPI schreiben
  //---------------------------------------------------
  MK3_SPI_SEND(wert);
 
  
  // Chip-Select auf Hi
  if((mode==ST7565_CMD1) || (mode==ST7565_DATA)) {    
    // nur wenn kein mehrbyte befehl
    ST7565_PORT|=ST7565_CS;
  }
}


//-----------------------------------------------------------------------------
// sendet ein Kommando an das Display
//-----------------------------------------------------------------------------
void ST7565_sendCommand(char befehl, char wert) { // interne Funktion 
  unsigned char b=0;
  
  switch(befehl) {
    case ST7565_DATA_TO_LCD : // sonderbefehl
      //-------------------------------------
      // das Byte wird an die akt. Adresse geschrieben,
      //-------------------------------------
      ST7565_send(ST7565_DATA,wert); // byte schreiben     
    ;break;
    case ST7565_SET_COLUMN_ADR_L : // 0000xxxx = 00h      
      wert=wert&0x0F;
      b=ST7565_SET_COLUMN_ADR_L|wert;
      ST7565_send(ST7565_CMD2,b);
    ;break; 
    case ST7565_SET_COLUMN_ADR_H : // 0001xxxx = 10h
      wert=wert&0x0F;
      b=ST7565_SET_COLUMN_ADR_H|wert;
      ST7565_send(ST7565_CMD1,b);
    ;break; 
    case ST7565_SET_PAGE_ADR : // 1011xxxx = B0h
      wert=wert&0x0F;
      ST7565_LCD.akt_page_adr=wert;
      b=ST7565_SET_PAGE_ADR|wert;
      ST7565_send(ST7565_CMD1,b);
    ;break; 
  }    
} 



/************************************************************************/
/* Filename: font90_8x5                                                 */
/* Version : 1.01                                                       */
/* Date    : 02.11.2011                                                 */
/* Owner   : UB                                                         */
/*----------------------------------------------------------------------*/
/* Remark  : Font-File zum LCD-Modul                                    */
/*           Schriftgroesse : 8x5 bit                                   */
/*                            wobei ein Zeichen aus 5x7 Pixel besteht   */
/*           Vorsicht : das Array muss fortlaufend sein...              */
/*                      z.B. von Ascii-Zeichen 32d (Space)              */
/*                           bis Ascii-Zeichen 126d (Tilde)             */
/*----------------------------------------------------------------------*/
/*                                                                      */
/*  z.B. Ascii 82d = R -> 0x7F,0x09,0x19,0x29,0x46,                     */
/*                                                                      */
/*                  (linke obere Ecke = LSB                             */
/*                                                                      */
/*                       12345 (Spalte)                                 */
/*                                                                      */
/*                  01h  ####:         = 0x7F (Spalte-1)                */
/*                  02h  #:::#         = 0x09 (Spalte-2)                */
/*                  04h  #:::#         = 0x19 (Spalte-3)                */
/*                  08h  ####:         = 0x29 (Spalte-4)                */
/*                  10h  #:#::         = 0x46 (Spalte-5)                */
/*                  20h  #::#:                                          */
/*                  40h  #:::#                                          */
/*                  80h  :::::                                          */
/*                                                                      */
/************************************************************************/

typedef struct {              // def von Font 8x5
  char b0;
  char b1;
  char b2;
  char b3;
  char b4;
}Font8x5_t;


// Zeichensatz wird im FLASH gespeichert

Font8x5_t font8x5[] UB_PROGMEM = {
  {0x00,0x00,0x00,0x00,0x00}, // Space
  {0x00,0x00,0x4F,0x00,0x00}, // !
  {0x00,0x07,0x00,0x07,0x00}, // "
  {0x14,0x7F,0x14,0x7F,0x14}, // #
  {0x24,0x2A,0x7F,0x2A,0x12}, // $
  {0x63,0x13,0x08,0x64,0x63}, // %
  {0x30,0x4E,0x5D,0x26,0x50}, // &
  {0x04,0x02,0x01,0x00,0x00}, // '
  {0x00,0x1C,0x22,0x41,0x00}, // (
  {0x00,0x41,0x22,0x1C,0x00}, // )
  {0x14,0x08,0x3E,0x08,0x14}, // *
  {0x08,0x08,0x3E,0x08,0x08}, // +
  {0xC0,0x30,0x00,0x00,0x00}, // ,
  {0x08,0x08,0x08,0x08,0x08}, // -
  {0x00,0x60,0x60,0x00,0x00}, // .
  {0x60,0x10,0x08,0x04,0x03}, // /
  {0x3E,0x41,0x41,0x3E,0x00}, // 0
  {0x00,0x42,0x7F,0x40,0x00}, // 1
  {0x72,0x49,0x49,0x49,0x46}, // 2
  {0x21,0x41,0x49,0x4D,0x33}, // 3
  {0x18,0x14,0x12,0x7F,0x10}, // 4
  {0x47,0x45,0x45,0x45,0x39}, // 5
  {0x38,0x4C,0x4A,0x49,0x30}, // 6
  {0x01,0x71,0x09,0x05,0x03}, // 7
  {0x36,0x49,0x49,0x49,0x36}, // 8
  {0x06,0x49,0x29,0x19,0x0E}, // 9
  {0x00,0x00,0x44,0x00,0x00}, // :
  {0x40,0x34,0x00,0x00,0x00}, // ;
  {0x08,0x14,0x22,0x41,0x41}, // <
  {0x14,0x14,0x14,0x14,0x14}, // =
  {0x41,0x41,0x22,0x14,0x08}, // >
  {0x02,0x01,0x51,0x09,0x06}, // ?
  {0x3E,0x41,0x5D,0x55,0x5E}, // @
  {0x7E,0x09,0x09,0x09,0x7E}, // A
  {0x7F,0x49,0x49,0x49,0x36}, // B
  {0x3E,0x41,0x41,0x41,0x22}, // C
  {0x7F,0x41,0x41,0x22,0x1C}, // D
  {0x7F,0x49,0x49,0x49,0x41}, // E
  {0x7F,0x09,0x09,0x09,0x01}, // F
  {0x3E,0x41,0x41,0x49,0x7A}, // G
  {0x7F,0x08,0x08,0x08,0x7F}, // H
  {0x00,0x41,0x7F,0x41,0x00}, // I
  {0x20,0x40,0x41,0x3F,0x01}, // J
  {0x7F,0x08,0x14,0x22,0x41}, // K
  {0x7F,0x40,0x40,0x40,0x40}, // L
  {0x7F,0x02,0x0C,0x02,0x7F}, // M
  {0x7F,0x06,0x08,0x30,0x7F}, // N
  {0x3E,0x41,0x41,0x41,0x3E}, // O
  {0x7F,0x09,0x09,0x09,0x06}, // P
  {0x3E,0x41,0x51,0x21,0x5E}, // Q
  {0x7F,0x09,0x19,0x29,0x46}, // R
  {0x46,0x49,0x49,0x49,0x31}, // S
  {0x01,0x01,0x7F,0x01,0x01}, // T
  {0x3F,0x40,0x40,0x40,0x3F}, // U
  {0x1F,0x20,0x40,0x20,0x1F}, // V
  {0x7F,0x20,0x18,0x20,0x7F}, // W
  {0x63,0x14,0x08,0x14,0x63}, // X
  {0x07,0x08,0x78,0x08,0x07}, // Y
  {0x61,0x51,0x49,0x45,0x43}, // Z
  {0x7F,0x41,0x41,0x00,0x00}, // [  
  {0x02,0x04,0x08,0x10,0x20}, // Backslash 
  {0x00,0x00,0x41,0x41,0x7F}, // ]
  {0x04,0x02,0x01,0x02,0x04}, // ^
  {0x80,0x80,0x80,0x80,0x80}, // _
  {0x00,0x01,0x02,0x04,0x00}, // `
  {0x38,0x44,0x44,0x28,0x7C}, // a
  {0x7F,0x28,0x44,0x44,0x38}, // b
  {0x38,0x44,0x44,0x44,0x00}, // c
  {0x38,0x44,0x44,0x28,0x7F}, // d
  {0x38,0x54,0x54,0x54,0x18}, // e
  {0x00,0x08,0x7E,0x09,0x00}, // f
  {0x18,0xA4,0xA4,0x98,0x7C}, // g
  {0x7F,0x08,0x04,0x04,0x78}, // h
  {0x00,0x44,0x7D,0x40,0x00}, // i
  {0x40,0x80,0x80,0x7D,0x00}, // j
  {0x7F,0x10,0x18,0x24,0x42}, // k
  {0x00,0x41,0x7F,0x40,0x00}, // l
  {0x7C,0x04,0x78,0x04,0x78}, // m
  {0x7C,0x04,0x04,0x78,0x00}, // n
  {0x38,0x44,0x44,0x44,0x38}, // o
  {0xFC,0x18,0x24,0x24,0x18}, // p
  {0x18,0x24,0x24,0x18,0xFC}, // q
  {0x7C,0x08,0x04,0x04,0x04}, // r
  {0x48,0x54,0x54,0x54,0x24}, // s
  {0x00,0x04,0x3F,0x44,0x00}, // t
  {0x3C,0x40,0x40,0x40,0x7C}, // u
  {0x0C,0x30,0x40,0x30,0x0C}, // v
  {0x3C,0x40,0x20,0x40,0x3C}, // w
  {0x44,0x28,0x10,0x28,0x44}, // x
  {0x1C,0xA0,0xA0,0xA0,0x7C}, // y
  {0x44,0x64,0x54,0x4C,0x44}, // z
  {0x08,0x36,0x41,0x00,0x00}, // {
  {0x00,0x00,0x77,0x00,0x00}, // |
  {0x00,0x00,0x41,0x36,0x08}, // }
  {0x10,0x08,0x08,0x08,0x04}  // ~
};

#define LCD_FIRST_CHAR 0x20 // erste Ascii-Nr im Array
#define LCD_LAST_CHAR  0x7E // letzt Ascii-Nr im Array


//-----------------------------------------------------------------------------
// zeichnet ein Ascii-Zeichen an Cursor-Pos
//-----------------------------------------------------------------------------
void LCD_ST7565_PRINT_ASC(unsigned char wert) { // zeichnet ein Ascii
  unsigned int c_pos;
  unsigned char flash_wert;

  if((wert<LCD_FIRST_CHAR) || (wert>LCD_LAST_CHAR)) {
    // wenn Ascii-Zeichen nicht erlaubt ist
    wert=LCD_FIRST_CHAR;
  }

  // position im Array berechnen
  c_pos=wert-LCD_FIRST_CHAR;
   
  // ein komplette Ascii-Zeichen (aus 5Bytes) zeichnen
  flash_wert=pgm_read_byte(&font8x5[c_pos].b0);
  LCD_ST7565_PRINT_BYTE(flash_wert);
  flash_wert=pgm_read_byte(&font8x5[c_pos].b1);
  LCD_ST7565_PRINT_BYTE(flash_wert);
  flash_wert=pgm_read_byte(&font8x5[c_pos].b2);
  LCD_ST7565_PRINT_BYTE(flash_wert);
  flash_wert=pgm_read_byte(&font8x5[c_pos].b3);
  LCD_ST7565_PRINT_BYTE(flash_wert);
  flash_wert=pgm_read_byte(&font8x5[c_pos].b4);
  LCD_ST7565_PRINT_BYTE(flash_wert);
  // abschliesend noch ein trennungsbyte
  LCD_ST7565_PRINT_BYTE(0x00);

}